@@ -10,7 +10,7 @@ def fermat(n):
     for x in count(1):
         for y in range(1, x+1):
             for z in range(1, x**n+y**n + 1):
-                from pudb import set_trace; set_trace()
+                #from pudb import set_trace; set_trace()
                 if x**n + y**n == z**n:
                     yield x, y, z
 
