@@ -1,11 +1,22 @@
 #! /usr/bin/env python
 # -*- coding: utf-8 -*-
 
-import urwid.raw_display
+try:
+    import urwid.curses_display as display
+except ImportError:
+    import urwid.raw_display as display
+
 import urwid
 import bdb
 from code import InteractiveConsole
 
+try:
+    import readline
+    import rlcompleter
+    HAVE_READLINE = True
+except ImportError:
+    HAVE_READLINE = False
+
 # TODO: Pop open local variables
 
 
@@ -181,7 +192,19 @@ class Debugger(bdb.Bdb):
 
 # UI stuff --------------------------------------------------------------------
 class MyConsole(InteractiveConsole):
-    pass
+    def __init__(self, locals):
+        InteractiveConsole.__init__(self, locals)
+
+        if HAVE_READLINE:
+            import os
+            import atexit
+
+            histfile = os.path.join(os.environ["HOME"], ".pudbhist")
+            if os.access(histfile, os.R_OK):
+                readline.read_history_file(histfile)
+            atexit.register(readline.write_history_file, histfile)
+            readline.parse_and_bind("tab: complete")
+
 
 
 
@@ -273,8 +296,10 @@ class SourceLine(urwid.FlowWidget):
         else:
             bp = " "
 
-        if focus or self.highlight:
+        if focus:
             attrs.append("focused")
+        elif self.highlight:
+            attrs.append("highlighted")
 
         if not attrs and self.attr is not None:
             attr = self.attr
@@ -567,7 +592,10 @@ class DebuggerUI(object):
             else:
                 banner = ""
 
-            loc = self.debugger.curframe.f_locals
+            curframe = self.debugger.curframe
+            loc = curframe.f_locals.copy()
+            loc.update(curframe.f_globals)
+
             cons = MyConsole(loc)
             cons.interact(banner)
             self.screen.start()
@@ -763,7 +791,7 @@ class DebuggerUI(object):
         self.top.listen("?", help)
 
         # setup ---------------------------------------------------------------
-        self.screen = urwid.raw_display.Screen()
+        self.screen = display.Screen()
         self.setup_palette(self.screen)
 
         self.show_count = 0
@@ -881,8 +909,10 @@ class DebuggerUI(object):
             # highlighting
             ("source", "yellow", "dark blue"),
             ("focused source", "black", "dark green"),
+            ("highlighted source", "black", "dark magenta"),
             ("current source", "black", "dark cyan"),
             ("current focused source", "white", "dark cyan"),
+            ("current highlighted source", "white", "dark cyan"),
 
             ("keyword", add_setting("white", "bold"), "dark blue"),
             ("literal", "light magenta", "dark blue"),
@@ -1059,16 +1089,22 @@ class DebuggerUI(object):
 
         loc_list = []
 
+        def format_value(v):
+            try:
+                return str(v)
+            except:
+                return "*** ERROR in str() ***"
+
         if "__return__" in vars:
             loc_list.append(
                     urwid.AttrWrap(
-                        SelectableText("Return: %s" % locals["__return__"],
+                        SelectableText("Return: %s" % format_value(locals["__return__"]),
                             wrap="clip"), 
                         "return value", "focused return value"))
 
         loc_list.extend(
                 urwid.AttrWrap(
-                    SelectableText("%s: %s" % (var, locals[var]),
+                    SelectableText("%s: %s" % (var, format_value(locals[var])),
                         wrap="clip"), 
                     None, "focused variable")
                 for var in vars
